# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from collections import defaultdict
import inspect
from itertools import chain
import types
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Callable,
    Dict,
    DefaultDict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from monkeytype.compat import (
    is_any,
    is_generic,
    is_generic_of,
    is_typed_dict,
    is_union,
    name_of_generic,
    types_equal,
)

from mypy_extensions import TypedDict


DUMMY_TYPED_DICT_NAME = 'DUMMY_NAME'
DUMMY_REQUIRED_TYPED_DICT_NAME = 'REQUIRED_TYPED_DICT_NAME'
DUMMY_OPTIONAL_TYPED_DICT_NAME = 'OPTIONAL_TYPED_DICT_NAME'


# Functions like shrink_types and get_type construct new types at runtime.
# Mypy cannot currently type these functions, so the type signatures for this
# file live in typing.pyi.


def make_typed_dict(*, required_fields=None, optional_fields=None) -> type:
    required_fields = required_fields or {}
    optional_fields = optional_fields or {}
    assert required_fields.keys().isdisjoint(optional_fields.keys())
    return TypedDict(DUMMY_TYPED_DICT_NAME, {
        "required_fields": TypedDict(DUMMY_REQUIRED_TYPED_DICT_NAME, required_fields),
        "optional_fields": TypedDict(DUMMY_OPTIONAL_TYPED_DICT_NAME, optional_fields)
    })


def field_annotations(typed_dict) -> Tuple[Dict[str, type], Dict[str, type]]:
    """Return the required and optional fields in the TypedDict."""
    return (typed_dict.__annotations__["required_fields"].__annotations__,
            typed_dict.__annotations__["optional_fields"].__annotations__)


def is_anonymous_typed_dict(typ: type) -> bool:
    """Return true if this is an anonymous TypedDict as generated by MonkeyType."""
    return is_typed_dict(typ) and typ.__name__ == DUMMY_TYPED_DICT_NAME


def shrink_typed_dict_types(typed_dicts: List[type], max_typed_dict_size: int) -> type:
    """Shrink a list of TypedDicts into one with the required fields and the optional fields.
    Required fields are keys that appear as a required field in all the TypedDicts.
    Optional fields are those that appear as a required field in only some
    of the TypedDicts or appear as a optional field in even one TypedDict.
    If the same key has multiple value types, then its value is the Union of the value types."""
    num_typed_dicts = len(typed_dicts)
    key_value_types_dict = defaultdict(list)
    existing_optional_fields = []
    for typed_dict in typed_dicts:
        required_fields, optional_fields = field_annotations(typed_dict)
        for key, value_type in required_fields.items():
            key_value_types_dict[key].append(value_type)
        existing_optional_fields.extend(optional_fields.items())

    required_fields = {key: value_types
                       for key, value_types in key_value_types_dict.items()
                       if len(value_types) == num_typed_dicts}
    optional_fields = defaultdict(list)
    for key, value_types in key_value_types_dict.items():
        if len(value_types) != num_typed_dicts:
            optional_fields[key] = value_types
    for key, value_type in existing_optional_fields:
        optional_fields[key].append(value_type)

    if len(required_fields) + len(optional_fields) > max_typed_dict_size:
        value_type = shrink_types(list(chain.from_iterable(chain(required_fields.values(),
                                                                 optional_fields.values()))),
                                  max_typed_dict_size)
        return Dict[str, value_type]
    required_fields = {key: shrink_types(list(value_types), max_typed_dict_size)
                       for key, value_types in required_fields.items()}
    optional_fields = {key: shrink_types(list(value_types), max_typed_dict_size)
                       for key, value_types in optional_fields.items()}
    return make_typed_dict(required_fields=required_fields, optional_fields=optional_fields)


def shrink_types(types, max_typed_dict_size):
    """Return the smallest type equivalent to Union[types].
    If all the types are anonymous TypedDicts, shrink them ourselves.
    Otherwise, recursively turn the anonymous TypedDicts into Dicts.
    Union will handle deduplicating types (both by equality and subtype relationships)."""
    types = tuple(types)
    if len(types) == 0:
        return Any
    if all(is_anonymous_typed_dict(typ) for typ in types):
        return shrink_typed_dict_types(types, max_typed_dict_size)
    # Don't rewrite anonymous TypedDict to Dict if the types are all the same,
    # such as List[TypedDict(...)].
    if all(types_equal(typ, types[0]) for typ in types[1:]):
        return types[0]
    all_dict_types = tuple(RewriteAnonymousTypedDictToDict().rewrite(typ) for typ in types)
    return Union[all_dict_types]


def make_iterator(typ):
    return Iterator[typ]


def make_generator(yield_typ, send_typ, return_typ):
    return Generator[yield_typ, send_typ, return_typ]


_BUILTIN_CALLABLE_TYPES = (
    types.FunctionType,
    types.LambdaType,
    types.MethodType,
    types.BuiltinMethodType,
    types.BuiltinFunctionType,
)


def get_dict_type(dct, max_typed_dict_size):
    """Return a TypedDict for `dct` if all the keys are strings.
    Else, default to the union of the keys and of the values."""
    if len(dct) == 0:
        # Special-case this because returning an empty TypedDict is
        # unintuitive, especially when you've "disabled" TypedDict generation
        # by setting `max_typed_dict_size` to 0.
        return Dict[Any, Any]
    if (all(isinstance(k, str) for k in dct.keys())
            and (max_typed_dict_size is None or len(dct) <= max_typed_dict_size)):
        return make_typed_dict(required_fields={k: get_type(v, max_typed_dict_size) for k, v in dct.items()})
    else:
        key_type = shrink_types((get_type(k, max_typed_dict_size) for k in dct.keys()), max_typed_dict_size)
        val_type = shrink_types((get_type(v, max_typed_dict_size) for v in dct.values()), max_typed_dict_size)
        return Dict[key_type, val_type]


def get_type(obj, max_typed_dict_size):
    """Return the static type that would be used in a type hint"""
    if isinstance(obj, type):
        return Type[obj]
    elif isinstance(obj, _BUILTIN_CALLABLE_TYPES):
        return Callable
    elif isinstance(obj, types.GeneratorType):
        return Iterator[Any]
    typ = type(obj)
    if typ is list:
        elem_type = shrink_types((get_type(e, max_typed_dict_size) for e in obj), max_typed_dict_size)
        return List[elem_type]
    elif typ is set:
        elem_type = shrink_types((get_type(e, max_typed_dict_size) for e in obj), max_typed_dict_size)
        return Set[elem_type]
    elif typ is dict:
        return get_dict_type(obj, max_typed_dict_size)
    elif typ is defaultdict:
        key_type = shrink_types((get_type(k, max_typed_dict_size) for k in obj.keys()), max_typed_dict_size)
        val_type = shrink_types((get_type(v, max_typed_dict_size) for v in obj.values()), max_typed_dict_size)
        return DefaultDict[key_type, val_type]
    elif typ is tuple:
        return Tuple[tuple(get_type(e, max_typed_dict_size) for e in obj)]
    return typ


NoneType = type(None)
NotImplementedType = type(NotImplemented)
mappingproxy = type(range.__dict__)


T = TypeVar("T")


class GenericTypeRewriter(Generic[T], ABC):
    @abstractmethod
    def make_builtin_tuple(self, elements): ...

    @abstractmethod
    def make_container_type(self, container_type, element): ...

    @abstractmethod
    def make_anonymous_typed_dict(self, required_fields, optional_fields): ...

    @abstractmethod
    def make_builtin_typed_dict(self, name, annotations, total): ...

    @abstractmethod
    def generic_rewrite(self, typ): ...

    @abstractmethod
    def rewrite_container_type(self, container_type): ...

    @abstractmethod
    def rewrite_malformed_container(self, container): ...

    @abstractmethod
    def rewrite_type_variable(self, type_variable): ...

    def _rewrite_container(self, cls, container):
        if container.__module__ != "typing":
            return self.rewrite_malformed_container(container)
        args = getattr(container, '__args__', None)
        if args is None:
            return self.rewrite_malformed_container(container)
        elif args == ((),):  # special case of empty tuple `Tuple[()]`
            elems = self.make_builtin_tuple(())
        else:
            elems = self.make_builtin_tuple(self.rewrite(elem) for elem in container.__args__)
        return self.make_container_type(self.rewrite_container_type(cls), elems)

    def rewrite_Dict(self, dct):
        return self._rewrite_container(Dict, dct)

    def rewrite_List(self, lst):
        return self._rewrite_container(List, lst)

    def rewrite_Set(self, st):
        return self._rewrite_container(Set, st)

    def rewrite_Tuple(self, tup):
        return self._rewrite_container(Tuple, tup)

    def rewrite_Generator(self, generator):
        return self._rewrite_container(Generator, generator)

    def rewrite_anonymous_TypedDict(self, typed_dict):
        assert is_anonymous_typed_dict(typed_dict)
        required_fields, optional_fields = field_annotations(typed_dict)
        return self.make_anonymous_typed_dict(required_fields={name: self.rewrite(typ)
                                                               for name, typ in required_fields.items()},
                                              optional_fields={name: self.rewrite(typ)
                                                               for name, typ in optional_fields.items()})

    def rewrite_TypedDict(self, typed_dict):
        if is_anonymous_typed_dict(typed_dict):
            return self.rewrite_anonymous_TypedDict(typed_dict)
        return self.make_builtin_typed_dict(typed_dict.__name__,
                                            {name: self.rewrite(typ)
                                             for name, typ in typed_dict.__annotations__.items()},
                                            total=typed_dict.__total__)

    def rewrite_Union(self, union):
        return self._rewrite_container(Union, union)

    def rewrite(self, typ):
        if is_any(typ):
            typname = 'Any'
        elif is_union(typ):
            typname = 'Union'
        elif is_typed_dict(typ):
            typname = 'TypedDict'
        elif is_generic(typ):
            typname = name_of_generic(typ)
        else:
            typname = getattr(typ, '__name__', None)
        rewriter = getattr(
            self, 'rewrite_' + typname, None) if typname else None
        if rewriter:
            return rewriter(typ)
        if isinstance(typ, TypeVar):
            return self.rewrite_type_variable(typ)
        return self.generic_rewrite(typ)


class TypeRewriter(GenericTypeRewriter[type]):
    """TypeRewriter provides a visitor for rewriting parts of types"""

    def make_anonymous_typed_dict(self, required_fields, optional_fields):
        return make_typed_dict(required_fields=required_fields, optional_fields=optional_fields)

    def make_builtin_typed_dict(self, name, annotations, total):
        return TypedDict(name, annotations, total=total)

    def generic_rewrite(self, typ):
        return typ

    def rewrite_container_type(self, container_type):
        return container_type

    def rewrite_malformed_container(self, container):
        return container

    def rewrite_type_variable(self, type_variable):
        return type_variable

    def make_builtin_tuple(self, elements):
        return tuple(elements)

    def make_container_type(self, container_type, element):
        return container_type[element]


class RemoveEmptyContainers(TypeRewriter):
    """Remove redundant, empty containers from union types.

    Empty containers are typed as C[Any] by MonkeyType. They should be removed
    if there is a single concrete, non-null type in the Union. For example,

        Union[Set[Any], Set[int]] -> Set[int]

    Union[] handles the case where there is only a single type left after
    removing the empty container.
    """

    def _is_empty(self, typ):
        args = getattr(typ, '__args__', [])
        return args and all(is_any(e) for e in args)

    def rewrite_Union(self, union):
        elems = tuple(
            self.rewrite(e) for e in union.__args__ if not self._is_empty(e))
        if elems:
            return Union[elems]
        return union


class RewriteConfigDict(TypeRewriter):
    """Union[Dict[K, V1], ..., Dict[K, VN]] -> Dict[K, Union[V1, ..., VN]]"""

    def rewrite_Union(self, union):
        key_type = None
        value_types = []
        for e in union.__args__:
            if not is_generic_of(e, Dict):
                return union
            key_type = key_type or e.__args__[0]
            if key_type != e.__args__[0]:
                return union
            value_types.extend(e.__args__[1:])
        return Dict[key_type, Union[tuple(value_types)]]


class RewriteLargeUnion(TypeRewriter):
    """Rewrite Union[T1, ..., TN] as Any for large N."""

    def __init__(self, max_union_len: int = 5):
        super().__init__()
        self.max_union_len = max_union_len

    def _rewrite_to_tuple(self, union):
        """Union[Tuple[V, ..., V], Tuple[V, ..., V], ...] -> Tuple[V, ...]"""
        value_type = None
        for t in union.__args__:
            if not is_generic_of(t, Tuple):
                return None
            value_type = value_type or t.__args__[0]
            if not all(vt is value_type for vt in t.__args__):
                return None
        return Tuple[value_type, ...]

    def rewrite_Union(self, union):
        if len(union.__args__) <= self.max_union_len:
            return union

        rw_union = self._rewrite_to_tuple(union)
        if rw_union is not None:
            return rw_union

        try:
            for ancestor in inspect.getmro(union.__args__[0]):
                if (
                    ancestor is not object and
                    all(issubclass(t, ancestor) for t in union.__args__)
                ):
                    return ancestor
        except (TypeError, AttributeError):
            pass
        return Any


class RewriteAnonymousTypedDictToDict(TypeRewriter):
    """TypedDict('Foo', {"k": v1, ...}) -> Dict[str, Union[v1, ...]]."""

    def rewrite_anonymous_TypedDict(self, typed_dict):
        assert is_anonymous_typed_dict(typed_dict)
        required_fields, optional_fields = field_annotations(typed_dict)
        all_value_types = [*required_fields.values(), *optional_fields.values()]
        if not all_value_types:
            # Special-case this because we can't justify any type.
            return Dict[Any, Any]
        return Dict[str, Union[tuple(self.rewrite(typ) for typ in all_value_types)]]


class ChainedRewriter(TypeRewriter):
    def __init__(self, rewriters: Iterable[TypeRewriter]) -> None:
        self.rewriters = rewriters

    def rewrite(self, typ):
        for rw in self.rewriters:
            typ = rw.rewrite(typ)
        return typ


class NoOpRewriter(TypeRewriter):
    def rewrite(self, typ):
        return typ


class RewriteGenerator(TypeRewriter):
    """Returns an Iterator, if the send_type and return_type of a Generator is None"""

    def rewrite_Generator(self, typ):
        args = typ.__args__
        if args[1] is NoneType and args[2] is NoneType:
            return Iterator[args[0]]
        return typ


DEFAULT_REWRITER = ChainedRewriter((
    RemoveEmptyContainers(),
    RewriteConfigDict(),
    RewriteLargeUnion(),
    RewriteGenerator(),
))



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from typing import Any
from mypy_extensions import _TypedDictMeta  # type: ignore


def is_typed_dict(typ: type) -> bool:
    """Test indirectly using _TypedDictMeta because TypedDict does not support `isinstance`."""
    return isinstance(typ, _TypedDictMeta)


try:
    # Python 3.7
    from typing import Union, _GenericAlias, ForwardRef  # type: ignore

    def is_any(typ: Any) -> bool:
        return typ is Any

    def is_union(typ: Any) -> bool:
        return typ is Union or is_generic(typ) and typ.__origin__ is Union

    try:
        # Python 3.9
        from typing import _SpecialGenericAlias

        def is_generic(typ: Any) -> bool:
            return typ is Union or isinstance(typ, _GenericAlias) or isinstance(typ, _SpecialGenericAlias)

    except ImportError:
        def is_generic(typ: Any) -> bool:
            return typ is Union or isinstance(typ, _GenericAlias)

    def is_generic_of(typ: Any, gen: Any) -> bool:
        return is_generic(typ) and typ.__origin__ is gen.__origin__

    def qualname_of_generic(typ: Any) -> str:
        return typ._name or typ.__origin__.__qualname__

    def name_of_generic(typ: Any) -> str:
        return typ._name or typ.__origin__.__name__

    def is_forward_ref(typ: Any) -> bool:
        return isinstance(typ, ForwardRef)

    def make_forward_ref(s: str) -> type:
        return ForwardRef(s)

    def repr_forward_ref() -> str:
        """For checking the test output when ForwardRef is printed."""
        return 'ForwardRef'

    def __are_typed_dict_types_equal(type1: type, type2: type) -> bool:
        """Return true if the two TypedDicts are equal.
        Doing this explicitly because
        TypedDict('Foo', {'a': int}) != TypedDict('Foo', {'a': int})."""

        if not is_typed_dict(type2):
            return False
        total1 = getattr(type1, "__total__", True)
        total2 = getattr(type2, "__total__", True)
        return (
            type1.__name__ == type2.__name__
            and total1 == total2
            and type1.__annotations__ == type2.__annotations__
        )

    def types_equal(typ: type, other_type: type) -> bool:
        return typ == other_type

except ImportError:
    # Python 3.6
    from typing import _Any, _Union, GenericMeta, _ForwardRef  # type: ignore

    def is_any(typ: Any) -> bool:
        return isinstance(typ, _Any)

    def is_union(typ: Any) -> bool:
        return isinstance(typ, _Union)

    def is_generic(typ: Any) -> bool:
        return isinstance(typ, (GenericMeta, _Union))

    def is_generic_of(typ: Any, gen: Any) -> bool:
        return issubclass(typ, gen)

    def qualname_of_generic(typ: Any) -> str:
        return typ.__qualname__

    def name_of_generic(typ: Any) -> str:
        return typ.__name__

    def is_forward_ref(typ: Any) -> bool:
        return isinstance(typ, _ForwardRef)

    def make_forward_ref(s: str) -> type:
        return _ForwardRef(s)

    def repr_forward_ref() -> str:
        """For checking the test output when ForwardRef is printed."""
        return '_ForwardRef'

    def __are_typed_dict_types_equal(type1: type, type2: type) -> bool:
        """Return true if the two TypedDicts are equal.
        Doing this explicitly because
        TypedDict('Foo', {'a': int}) != TypedDict('Foo', {'a': int})."""

        if not is_typed_dict(type2):
            return False
        total1 = getattr(type1, "__total__", True)
        total2 = getattr(type2, "__total__", True)
        return (type1.__name__ == type2.__name__
                and total1 == total2
                and type1.__annotations__.keys() == type2.__annotations__.keys()
                and all(types_equal(type1.__annotations__[key], type2.__annotations__[key])
                        for key in type1.__annotations__.keys()))

    def types_equal(typ: type, other_type: type) -> bool:
        # Types for which equality with inner TypedDict doesn't work correctly on 3.6.
        special_container_types = ["List", "Dict", "Tuple", "Set", "DefaultDict"]

        if (
            (is_any(typ) and is_any(other_type))
            or (is_union(typ) and is_union(other_type))
            or (is_typed_dict(typ) and is_typed_dict(other_type))
        ):
            pass
        elif is_generic(typ) and is_generic(other_type):
            if (
                name_of_generic(typ) in special_container_types
                and name_of_generic(other_type) in special_container_types
            ):
                args = getattr(typ, "__args__", [])
                other_args = getattr(other_type, "__args__", [])
                return (
                    name_of_generic(typ) == name_of_generic(other_type)
                    and len(args) == len(other_args)
                    and all(types_equal(arg_type, other_arg_type)
                            for arg_type, other_arg_type in zip(args, other_args))
                )
        return typ == other_type


# HACK: MonkeyType monkey-patches _TypedDictMeta!
# We need this to compare TypedDicts recursively.
_TypedDictMeta.__eq__ = __are_typed_dict_types_equal



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import datetime
import logging
import sqlite3

from typing import (
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from monkeytype.db.base import (
    CallTraceStore,
    CallTraceThunk,
)
from monkeytype.encoding import (
    CallTraceRow,
    serialize_traces,
)
from monkeytype.tracing import CallTrace

logger = logging.getLogger(__name__)


DEFAULT_TABLE = 'monkeytype_call_traces'


def create_call_trace_table(conn: sqlite3.Connection, table: str = DEFAULT_TABLE) -> None:
    query = """
CREATE TABLE IF NOT EXISTS {table} (
  created_at  TEXT,
  module      TEXT,
  qualname    TEXT,
  arg_types   TEXT,
  return_type TEXT,
  yield_type  TEXT);
""".format(table=table)
    with conn:
        conn.execute(query)


QueryValue = Union[str, int]
ParameterizedQuery = Tuple[str, List[QueryValue]]


def make_query(table: str, module: str, qualname: Optional[str], limit: int) -> ParameterizedQuery:
    raw_query = """
    SELECT
        module, qualname, arg_types, return_type, yield_type
    FROM {table}
    WHERE
        module == ?
    """.format(table=table)
    values: List[QueryValue] = [module]
    if qualname is not None:
        raw_query += " AND qualname LIKE ? || '%'"
        values.append(qualname)
    raw_query += """
    GROUP BY
        module, qualname, arg_types, return_type, yield_type
    ORDER BY date(created_at) DESC
    LIMIT ?
    """
    values.append(limit)
    return raw_query, values


class SQLiteStore(CallTraceStore):
    def __init__(self, conn: sqlite3.Connection, table: str = DEFAULT_TABLE) -> None:
        self.conn = conn
        self.table = table

    @classmethod
    def make_store(cls, connection_string: str) -> 'CallTraceStore':
        conn = sqlite3.connect(connection_string)
        create_call_trace_table(conn)
        return cls(conn)

    def add(self, traces: Iterable[CallTrace]) -> None:
        values = []
        for row in serialize_traces(traces):
            values.append((datetime.datetime.now(), row.module, row.qualname,
                           row.arg_types, row.return_type, row.yield_type))
        with self.conn:
            self.conn.executemany(
                'INSERT INTO {table} VALUES (?, ?, ?, ?, ?, ?)'.format(table=self.table),
                values
            )

    def filter(
        self,
        module: str,
        qualname_prefix: Optional[str] = None,
        limit: int = 2000
    ) -> List[CallTraceThunk]:
        sql_query, values = make_query(self.table, module, qualname_prefix, limit)
        with self.conn:
            cur = self.conn.cursor()
            cur.execute(sql_query, values)
            return [CallTraceRow(*row) for row in cur.fetchall()]

    def list_modules(self) -> List[str]:
        with self.conn:
            cur = self.conn.cursor()
            cur.execute("""
                        SELECT module FROM {table}
                        GROUP BY module
                        ORDER BY date(created_at) DESC
                        """.format(table=self.table))
            return [row[0] for row in cur.fetchall() if row[0]]



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import json
import logging

from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Type,
    TypeVar,
)

from monkeytype.compat import is_any, is_union, is_generic, qualname_of_generic
from monkeytype.db.base import CallTraceThunk
from monkeytype.exceptions import InvalidTypeError
from monkeytype.tracing import CallTrace
from monkeytype.typing import NoneType, NotImplementedType, is_typed_dict, mappingproxy
from monkeytype.util import (
    get_func_in_module,
    get_name_in_module,
)
from mypy_extensions import TypedDict


logger = logging.getLogger(__name__)


# Types are converted to dictionaries of the following form before
# being JSON encoded and sent to storage:
#
#     {
#         'module': '<module>',
#         'qualname': '<qualname>',
#         'elem_types': [type_dict],
#     }
#
# The corresponding type alias should actually be
#
#     TypeDict = Dict[str, Union[str, TypeDict]]
#
# (or better, a TypedDict) but mypy does not support recursive type aliases:
#  https://github.com/python/mypy/issues/731
TypeDict = Dict[str, Any]


def typed_dict_to_dict(typ: type) -> TypeDict:
    elem_types: Dict[str, Any] = {}
    for k, v in typ.__annotations__.items():
        elem_types[k] = type_to_dict(v)
    return {
        'module': typ.__module__,
        'qualname': typ.__qualname__,
        'elem_types': elem_types,
        'is_typed_dict': True,
    }


def type_to_dict(typ: type) -> TypeDict:
    """Convert a type into a dictionary representation that we can store.

    The dictionary must:
        1. Be encodable as JSON
        2. Contain enough information to let us reify the type
    """
    if is_typed_dict(typ):
        return typed_dict_to_dict(typ)

    # Union and Any are special cases that aren't actually types.
    if is_union(typ):
        qualname = 'Union'
    elif is_any(typ):
        qualname = 'Any'
    elif is_generic(typ):
        qualname = qualname_of_generic(typ)
    else:
        qualname = typ.__qualname__
    d: TypeDict = {
        'module': typ.__module__,
        'qualname': qualname,
    }
    elem_types = getattr(typ, '__args__', None)
    if elem_types and is_generic(typ):
        # empty typing.Tuple is weird; the spec says it should be Tuple[()],
        # which results in __args__ of `((),)`
        if elem_types == ((),):
            elem_types = ()
        d['elem_types'] = [type_to_dict(t) for t in elem_types]
    return d


_HIDDEN_BUILTIN_TYPES: Dict[str, type] = {
    # Types that are inaccessible by their names in the builtins module.
    'NoneType': NoneType,
    'NotImplementedType': NotImplementedType,
    'mappingproxy': mappingproxy,
}


def typed_dict_from_dict(d: TypeDict) -> type:
    return TypedDict(d['qualname'],
                     {k: type_from_dict(v)
                      for k, v in d['elem_types'].items()})


def type_from_dict(d: TypeDict) -> type:
    """Given a dictionary produced by type_to_dict, return the equivalent type.

    Raises:
        NameLookupError if we can't reify the specified type
        InvalidTypeError if the named type isn't actually a type
    """
    module, qualname = d['module'], d['qualname']
    if d.get('is_typed_dict', False):
        return typed_dict_from_dict(d)
    if module == 'builtins' and qualname in _HIDDEN_BUILTIN_TYPES:
        typ = _HIDDEN_BUILTIN_TYPES[qualname]
    else:
        typ = get_name_in_module(module, qualname)
    if not (
        isinstance(typ, type) or
        is_any(typ) or
        is_generic(typ)
    ):
        raise InvalidTypeError(
            f"Attribute specified by '{qualname}' in module '{module}' "
            f"is of type {type(typ)}, not type."
        )
    elem_type_dicts = d.get('elem_types')
    if elem_type_dicts is not None and is_generic(typ):
        elem_types = tuple(type_from_dict(e) for e in elem_type_dicts)
        # mypy complains that a value of type `type` isn't indexable. That's
        # true, but we know typ is a subtype that is indexable. Even checking
        # with hasattr(typ, '__getitem__') doesn't help
        typ = typ[elem_types]  # type: ignore
    return typ


def type_to_json(typ: type) -> str:
    """Encode the supplied type as json using type_to_dict."""
    type_dict = type_to_dict(typ)
    return json.dumps(type_dict, sort_keys=True)


def type_from_json(typ_json: str) -> type:
    """Reify a type from the format produced by type_to_json."""
    type_dict = json.loads(typ_json)
    return type_from_dict(type_dict)


def arg_types_to_json(arg_types: Dict[str, type]) -> str:
    """Encode the supplied argument types as json"""
    type_dict = {name: type_to_dict(typ) for name, typ in arg_types.items()}
    return json.dumps(type_dict, sort_keys=True)


def arg_types_from_json(arg_types_json: str) -> Dict[str, type]:
    """Reify the encoded argument types from the format produced by arg_types_to_json."""
    arg_types = json.loads(arg_types_json)
    return {name: type_from_dict(type_dict) for name, type_dict in arg_types.items()}


TypeEncoder = Callable[[type], str]


def maybe_encode_type(encode: TypeEncoder, typ: Optional[type]) -> Optional[str]:
    if typ is None:
        return None
    return encode(typ)


TypeDecoder = Callable[[str], type]


def maybe_decode_type(decode: TypeDecoder, encoded: Optional[str]) -> Optional[type]:
    if (encoded is None) or (encoded == 'null'):
        return None
    return decode(encoded)


CallTraceRowT = TypeVar('CallTraceRowT', bound='CallTraceRow')


class CallTraceRow(CallTraceThunk):
    """A semi-structured call trace where each field has been json encoded."""

    def __init__(
        self,
        module: str,
        qualname: str,
        arg_types: str,
        return_type: Optional[str],
        yield_type: Optional[str]
    ) -> None:
        self.module = module
        self.qualname = qualname
        self.arg_types = arg_types
        self.return_type = return_type
        self.yield_type = yield_type

    @classmethod
    def from_trace(cls: Type[CallTraceRowT], trace: CallTrace) -> CallTraceRowT:
        module = trace.func.__module__
        qualname = trace.func.__qualname__
        arg_types = arg_types_to_json(trace.arg_types)
        return_type = maybe_encode_type(type_to_json, trace.return_type)
        yield_type = maybe_encode_type(type_to_json, trace.yield_type)
        return cls(module, qualname, arg_types, return_type, yield_type)

    def to_trace(self) -> CallTrace:
        function = get_func_in_module(self.module, self.qualname)
        arg_types = arg_types_from_json(self.arg_types)
        return_type = maybe_decode_type(type_from_json, self.return_type)
        yield_type = maybe_decode_type(type_from_json, self.yield_type)
        return CallTrace(function, arg_types, return_type, yield_type)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CallTraceRow):
            return (
                self.module,
                self.qualname,
                self.arg_types,
                self.return_type,
                self.yield_type,
            ) == (
                other.module,
                other.qualname,
                other.arg_types,
                other.return_type,
                other.yield_type,
            )
        return NotImplemented


def serialize_traces(traces: Iterable[CallTrace]) -> Iterable[CallTraceRow]:
    """Serialize an iterable of CallTraces to an iterable of CallTraceRow.

    Catches and logs exceptions, so a failure to serialize one CallTrace doesn't
    lose all traces.

    """
    for trace in traces:
        try:
            yield CallTraceRow.from_trace(trace)
        except Exception:
            logger.exception("Failed to serialize trace")



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from typing import (
    ContextManager,
    Optional,
)

from monkeytype.config import (
    Config,
    get_default_config,
)
from monkeytype.tracing import trace_calls

__version__ = "20.5.0"


def trace(config: Optional[Config] = None) -> ContextManager:
    """Context manager to trace and log all calls.

    Simple wrapper around `monkeytype.tracing.trace_calls` that uses trace
    logger, code filter, and sample rate from given (or default) config.
    """
    if config is None:
        config = get_default_config()
    return trace_calls(
        logger=config.trace_logger(),
        code_filter=config.code_filter(),
        sample_rate=config.sample_rate(),
        max_typed_dict_size=config.max_typed_dict_size(),
    )



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import (
    Iterable,
    List,
    Optional,
)


from monkeytype.tracing import CallTrace, CallTraceLogger


class CallTraceThunk(metaclass=ABCMeta):
    """A deferred computation that produces a CallTrace or raises an error."""

    @abstractmethod
    def to_trace(self) -> CallTrace:
        """Produces the CallTrace."""
        pass


class CallTraceStore(metaclass=ABCMeta):
    """An interface that all concrete calltrace storage backends must implement."""

    @abstractmethod
    def add(self, traces: Iterable[CallTrace]) -> None:
        """Store the supplied call traces in the backing store"""
        pass

    @abstractmethod
    def filter(
        self,
        module: str,
        qualname_prefix: Optional[str] = None,
        limit: int = 2000
    ) -> List[CallTraceThunk]:
        """Query the backing store for any traces that match the supplied query.

        By returning a list of thunks we let the caller get a partial result in the
        event that decoding one or more call traces fails.
        """
        pass

    @classmethod
    def make_store(cls, connection_string: str) -> 'CallTraceStore':
        """Create a new store instance.

        This is a factory function that is intended to be used by the CLI.
        """
        pass

    def list_modules(self) -> List[str]:
        """List of traced modules from the backing store"""
        raise NotImplementedError(
            f"Your CallTraceStore ({self.__class__.__module__}.{self.__class__.__name__}) "
            f"does not implement list_modules()"
        )


class CallTraceStoreLogger(CallTraceLogger):
    """A CallTraceLogger that stores logged traces in a CallTraceStore."""
    def __init__(self, store: CallTraceStore) -> None:
        self.store = store
        self.traces: List[CallTrace] = []

    def log(self, trace: CallTrace) -> None:
        if not trace.func.__module__ == '__main__':
            self.traces.append(trace)

    def flush(self) -> None:
        self.store.add(self.traces)
        self.traces = []



# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import importlib
import inspect
import types
import re

try:
    from django.utils.functional import cached_property  # type: ignore
except ImportError:
    cached_property = None

from typing import (
    Any,
    Callable,
)

from monkeytype.exceptions import (
    InvalidTypeError,
    NameLookupError,
)


def get_func_fqname(func: Callable) -> str:
    """Return the fully qualified function name."""
    return func.__module__ + '.' + func.__qualname__


def get_func_in_module(module: str, qualname: str) -> Callable:
    """Return the function specified by qualname in module.

    Raises:
        NameLookupError if we can't find the named function
        InvalidTypeError if we the name isn't a function
    """
    func = get_name_in_module(module, qualname)
    func = inspect.unwrap(func)
    if isinstance(func, types.MethodType):
        func = func.__func__
    elif isinstance(func, property):
        if func.fget is not None:
            if (func.fset is None) and (func.fdel is None):
                func = func.fget
            else:
                raise InvalidTypeError(
                    f"Property {module}.{qualname} has setter or deleter.")
        else:
            raise InvalidTypeError(
                f"Property {module}.{qualname} is missing getter")
    elif cached_property and isinstance(func, cached_property):
        func = func.func
    elif not isinstance(func, (types.FunctionType, types.BuiltinFunctionType)):
        raise InvalidTypeError(
            f"{module}.{qualname} is of type '{type(func)}', not function.")
    return func


def get_name_in_module(
    module: str,
    qualname: str,
    attr_getter: Callable[[Any, str], Any] = None,
) -> Any:
    """Return the python object specified by qualname in module.

    Raises:
        NameLookupError if the module/qualname cannot be retrieved.
    """
    if attr_getter is None:
        attr_getter = getattr
    try:
        obj = importlib.import_module(module)
    except ModuleNotFoundError:
        raise NameLookupError("No module named '%s'" % (module,))
    walked = []
    for part in qualname.split('.'):
        walked.append(part)
        try:
            obj = attr_getter(obj, part)
        except AttributeError:
            raise NameLookupError(
                "Module '%s' has no attribute '%s'"
                % (module, '.'.join(walked))
            )
    return obj


def pascal_case(s: str) -> str:
    return ''.join(a[0].upper() + a[1:] for a in re.split('([^a-zA-Z0-9])', s) if a.isalnum())
